/*************************************************************/
/*************************************************************/
/*                                                                  */
/*    Programari MPI                                                */
/*    Nom del programari:   Latency                                 */
/*    Descripció:   Test de latència entre node 0 i tota la resta   */
/*                  100.000 repeticions                             */
/*                                                                  */
/*    Estudiant:   José Antonio Martín Pérez                        */
/*    email:   jmartinperez1@uoc.edu  /  jamartin@protonmail.ch     */
/*                                                                  */
/*************************************************************/
/*************************************************************/


#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#define	NUMBER_REPS	100000     /* Nombre de repeticions  */
#define NODES           64


int main (int argc, char *argv[])
{
int reps,                   /* Nombre repecions per test */
    tag,                    /* MPI message tag parameter */
    numtasks,               /* Nombre de tasques MPI */
    rank,                   /* Identificador tasca MPI */
    dest, source,           /* Identificador nodes origen/destí */
    avgT,                   /* Temps promig per repetició (microsegons) */
    rc,                     /* Codi retorn */
    i,
    n;
double T1, T2, T3, T4,      /* Temps inici/fi per repetició */
    sumT,                   /* Suma de tots el temps */
    total_time,
    deltaT;                 /* Temps per una repetició */
char msg;                   /* Missatge d'un byte */
MPI_Status status; 

MPI_Init(&argc,&argv);
MPI_Comm_size(MPI_COMM_WORLD,&numtasks);
MPI_Comm_rank(MPI_COMM_WORLD,&rank);
MPI_Barrier(MPI_COMM_WORLD);

sumT = 0;
msg = 'x';
tag = 1;
reps = NUMBER_REPS;


/*******************/
/*      Node 0       */
/*******************/

if (rank == 0) {

   printf("Inici del test de latencia.  Nombre de repeticions = %d.\n", reps);
   printf("***************************************************\n");
   printf("Rep#       T1               T2            deltaT\n");
   dest = 1;
   source = 1;

   T3 = MPI_Wtime();
   
   for (n = 1; n <= reps; n++) {
      T1 = MPI_Wtime();     /* Marca d'inici del test */

      /* Enviament al node destí d'un byte */
      for (i = 1; i < NODES; i++)
      {
         rc = MPI_Send(&msg, 1, MPI_BYTE, i, tag, MPI_COMM_WORLD);
         if (rc != MPI_SUCCESS) {
            printf("Error enviament en tasca 0!\n");
            MPI_Abort(MPI_COMM_WORLD, rc);
            exit(1);
         }
       }

      for (i = 1; i < NODES; i++)
      {
         /* Esperem rebre resposta del node */
         rc = MPI_Recv(&msg, 1, MPI_BYTE, i, tag, MPI_COMM_WORLD, 
                    &status);
         if (rc != MPI_SUCCESS) {
            printf("Error recepcio en tasca 0!\n");
            MPI_Abort(MPI_COMM_WORLD, rc);
            exit(1);
         }      
      }

      /* calculate round trip time and print */
      T2 = MPI_Wtime();
      deltaT = T2 - T1;
      sumT += deltaT;

   }

   T4 = MPI_Wtime();
   total_time = T4 - T3;
   printf("\n\n*** Temps total = %8.8f microsegons \n\n", total_time);
} 


/*************************/
/*      Resta de nodes      */
/*************************/

else if (rank > 0) {
   dest = 0;
   source = 0;
   for (n = 1; n <= reps; n++) {
      rc = MPI_Recv(&msg, 1, MPI_BYTE, source, tag, MPI_COMM_WORLD, 
                    &status);
      if (rc != MPI_SUCCESS) {
         printf("Error de recepcio en tasca 1!\n");
         MPI_Abort(MPI_COMM_WORLD, rc);
         exit(1);
      }
      rc = MPI_Send(&msg, 1, MPI_BYTE, dest, tag, MPI_COMM_WORLD);
      if (rc != MPI_SUCCESS) {
         printf("Error enviament en tasca 1!\n");
         MPI_Abort(MPI_COMM_WORLD, rc);
         exit(1);
      }
   }

}

MPI_Finalize(); 

}

