/***********************************************************/
/***********************************************************/
/*                                                                */
/*    Programari MPI                                              */
/*    Nom del programari:   Array Assignment                      */
/*    Descripció:   Descomposició d'un vector entre nodes         */
/*                  10 repeticions                                */
/*                                                                */
/*    Estudiant:   José Antonio Martín Pérez                      */
/*    email:   jmartinperez1@uoc.edu  /  jamartin@protonmail.ch   */
/*                                                                */
/***********************************************************/
/***********************************************************/


#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#define  ARRAYSIZE	960000000      /* Mida de l'array a tractar */
#define  MASTER		0

#define ROUNDS  10       /* Nombre de repeticions */


float  data[ARRAYSIZE];

int main (int argc, char *argv[])
{
int   numtasks, taskid, rc, dest, offset, i, j, tag1,
      tag2, source, chunksize; 
int index;
float mysum, sum;
float update(int myoffset, int chunk, int myid);
MPI_Status status;

double T1, T2, total_time;

MPI_Init(&argc, &argv);
MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
if (numtasks % 4 != 0) {
   printf("Es requereix un nombre de tasques multiple de 4. Sortint...\n");
   MPI_Abort(MPI_COMM_WORLD, rc);
   exit(0);
   }
MPI_Comm_rank(MPI_COMM_WORLD,&taskid);
chunksize = (ARRAYSIZE / numtasks);
tag2 = 1;
tag1 = 2;



/****************/
/*     Node 0     */
/****************/

if (taskid == MASTER){

  T1 = MPI_Wtime();

  for (index=0; index<ROUNDS; index++) {

  /* Inicialització del vector */
  sum = 0;
  for(i=0; i<ARRAYSIZE; i++) {
    data[i] =  i * 1.0;
    sum = sum + data[i];
    }

  /* Enviament a cada node de la seva porció del vector. Node 0 es queda la primera d'aquestes */
  offset = chunksize;
  for (dest=1; dest<numtasks; dest++) {
    MPI_Send(&offset, 1, MPI_INT, dest, tag1, MPI_COMM_WORLD);
    MPI_Send(&data[offset], chunksize, MPI_FLOAT, dest, tag2, MPI_COMM_WORLD);
    offset = offset + chunksize;
    }

  /* Node 0 fa la seva part del treball */
  offset = 0;
  mysum = update(offset, chunksize, taskid);

  /* Rebem resultat de cada node */
  for (i=1; i<numtasks; i++) {
    source = i;
    MPI_Recv(&offset, 1, MPI_INT, source, tag1, MPI_COMM_WORLD, &status);
    MPI_Recv(&data[offset], chunksize, MPI_FLOAT, source, tag2,
      MPI_COMM_WORLD, &status);
    }

  /* Rebem suma final */  
  MPI_Reduce(&mysum, &sum, 1, MPI_FLOAT, MPI_SUM, MASTER, MPI_COMM_WORLD);
  offset = 0;
  }

  T2=MPI_Wtime();
  total_time = T2 - T1;
  printf("\n\n Temps total:  %8.8f \n\n" , total_time );

} 



/***********************/
/*     Resta de nodes     */
/***********************/

if (taskid > MASTER) {

 for (index=0; index<ROUNDS; index++) {

  /* Cada node rep la seva porció del vector */
  source = MASTER;
  MPI_Recv(&offset, 1, MPI_INT, source, tag1, MPI_COMM_WORLD, &status);
  MPI_Recv(&data[offset], chunksize, MPI_FLOAT, source, tag2, 
    MPI_COMM_WORLD, &status);

  mysum = update(offset, chunksize, taskid);

  /* Enviament del resultat computat al node 0 */
  dest = MASTER;
  MPI_Send(&offset, 1, MPI_INT, dest, tag1, MPI_COMM_WORLD);
  MPI_Send(&data[offset], chunksize, MPI_FLOAT, MASTER, tag2, MPI_COMM_WORLD);

  MPI_Reduce(&mysum, &sum, 1, MPI_FLOAT, MPI_SUM, MASTER, MPI_COMM_WORLD);

  }
}

MPI_Finalize();

}



//  Tasca per computar la suma valors de la porció del vector rebuda
//  ----------------------------------------------------------------

float update(int myoffset, int chunk, int myid) {
  int i; 
  float mysum;
  mysum = 0;
  for(i=myoffset; i < myoffset + chunk; i++) {
    data[i] = data[i] + i * 1.0;
    mysum = mysum + data[i];
    }
  return(mysum);
}


