/***********************************************************/
/***********************************************************/
/*                                                                */
/*    Programari MPI                                              */
/*    Nom del programari:   Matrix Multiply                       */
/*    Descripció:   Multiplicació de dues matrius grans           */
/*                  1.000 repeticions                             */
/*                                                                */
/*    Estudiant:   José Antonio Martín Pérez                      */
/*    email:   jmartinperez1@uoc.edu  /  jamartin@protonmail.ch   */
/*                                                                */
/***********************************************************/
/***********************************************************/


#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define NRA 248                /* Nombre files matriu A */
#define NCA 30                 /* Nombre columnes matriu A */
#define NCB 14                 /* Nombre columnes matriu B */
#define MASTER 0               /* Identificador primera tasca (0) */
#define FROM_MASTER 1          
#define FROM_WORKER 2          

#define ROUNDS 1000            /*  Nombre de repeticions */


int main (int argc, char *argv[])
{
int	numtasks,              /* Nombre de tasques */
	taskid,                /* Identificador tasca */
	numworkers,            /*  */
	source,                /* task id of message source */
	dest,                  /* task id of message destination */
	mtype,                 /* message type */
	rows,                  /* rows of matrix A sent to each worker */
	averow, extra, offset, /* used to determine rows sent to each worker */
	i, j, k, rc;           /* misc */
double	T1, T2, total_time,
        a[NRA][NCA],           /* matrix A to be multiplied */
	b[NCA][NCB],           /* matrix B to be multiplied */
	c[NRA][NCB];           /* result matrix C */
MPI_Status status;

MPI_Init(&argc,&argv);
MPI_Comm_rank(MPI_COMM_WORLD,&taskid);
MPI_Comm_size(MPI_COMM_WORLD,&numtasks);
if (numtasks < 2 ) {
  printf("Need at least two MPI tasks. Quitting...\n");
  MPI_Abort(MPI_COMM_WORLD, rc);
  exit(1);
  }
numworkers = numtasks-1;


/********************   Node 0 - Inici del procés   ********************/
   if (taskid == MASTER)
   {

    T1 = MPI_Wtime();   // Marca temps inicial

    int index;
    for (index=0; index<ROUNDS; index++)  {  

      // Inicialització de les matrius
      for (i=0; i<NRA; i++)
         for (j=0; j<NCA; j++)
            a[i][j]= i+j;
      for (i=0; i<NCA; i++)
         for (j=0; j<NCB; j++)
            b[i][j]= i*j;

      /* Enviament de les dades de la matriu a cada node */
      averow = NRA/numworkers;
      extra = NRA%numworkers;
      offset = 0;
      mtype = FROM_MASTER;
      for (dest=1; dest<=numworkers; dest++)
      {
         rows = (dest <= extra) ? averow+1 : averow;   	
         MPI_Send(&offset, 1, MPI_INT, dest, mtype, MPI_COMM_WORLD);
         MPI_Send(&rows, 1, MPI_INT, dest, mtype, MPI_COMM_WORLD);
         MPI_Send(&a[offset][0], rows*NCA, MPI_DOUBLE, dest, mtype,
                   MPI_COMM_WORLD);
         MPI_Send(&b, NCA*NCB, MPI_DOUBLE, dest, mtype, MPI_COMM_WORLD);
         offset = offset + rows;
      }

      /* Rebem les dades de cada node */
      mtype = FROM_WORKER;
      for (i=1; i<=numworkers; i++)
      {
         source = i;
         MPI_Recv(&offset, 1, MPI_INT, source, mtype, MPI_COMM_WORLD, &status);
         MPI_Recv(&rows, 1, MPI_INT, source, mtype, MPI_COMM_WORLD, &status);
         MPI_Recv(&c[offset][0], rows*NCB, MPI_DOUBLE, source, mtype, 
                  MPI_COMM_WORLD, &status);
      }

    }

      T2=MPI_Wtime();   // Marca temps final
      total_time = T2 - T1;
      printf("\n\n Temps total:  %8.8f \n\n" , total_time );

   }


/********************   Resta de nodes   ********************/
   if (taskid > MASTER)
   {
    int index;
    for (index=0; index<ROUNDS; index++)  {

      // Recepció de dades a computar
      mtype = FROM_MASTER;
      MPI_Recv(&offset, 1, MPI_INT, MASTER, mtype, MPI_COMM_WORLD, &status);
      MPI_Recv(&rows, 1, MPI_INT, MASTER, mtype, MPI_COMM_WORLD, &status);
      MPI_Recv(&a, rows*NCA, MPI_DOUBLE, MASTER, mtype, MPI_COMM_WORLD, &status);
      MPI_Recv(&b, NCA*NCB, MPI_DOUBLE, MASTER, mtype, MPI_COMM_WORLD, &status);

      // Comput de la seva part de la matriu
      for (k=0; k<NCB; k++)
         for (i=0; i<rows; i++)
         {
            c[i][k] = 0.0;
            for (j=0; j<NCA; j++)
               c[i][k] = c[i][k] + a[i][j] * b[j][k];
         }

      // Enviament de les dades computades al node 0
      mtype = FROM_WORKER;
      MPI_Send(&offset, 1, MPI_INT, MASTER, mtype, MPI_COMM_WORLD);
      MPI_Send(&rows, 1, MPI_INT, MASTER, mtype, MPI_COMM_WORLD);
      MPI_Send(&c, rows*NCB, MPI_DOUBLE, MASTER, mtype, MPI_COMM_WORLD);
    }
}

   MPI_Finalize();

}

