/***********************************************************/
/***********************************************************/
/*                                                                */
/*    Programari MPI                                              */
/*    Nom del programari:   Prime Generator                       */
/*    Descripció:   Generador de nombres primers                  */
/*                  100.000.000 nombres primers a generar         */
/*                                                                */
/*    Estudiant:   José Antonio Martín Pérez                      */
/*    email:   jmartinperez1@uoc.edu  /  jamartin@protonmail.ch   */
/*                                                                */
/***********************************************************/
/***********************************************************/


#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define LIMIT     100000000     /*  Nombre de primers a generar  */
#define FIRST     0             /*  Identificador de la tasca 0  */

int isprime(int n) {
int i,squareroot;
if (n>10) {
   squareroot = (int) sqrt(n);
   for (i=3; i<=squareroot; i=i+2)
      if ((n%i)==0)
         return 0;
   return 1;
   }
else
   return 0;
}


int main (int argc, char *argv[])
{
int   ntasks,               /* Nombre total de tasques */
      rank,                 /* Identificador de tasca */
      n,                    /* Variable per a bucles */
      pc,                   /* Comptador de nombres primers */
      pcsum,                /* Nombre de primers trobats per totes les tasques */
      foundone,             /* Primer més recenment trobat */
      maxprime,             /* Primer més gran trobat */
      mystart,              /* On comença a calcular cada tasca */
      stride;               

double start_time,end_time;
	
MPI_Init(&argc,&argv);
MPI_Comm_rank(MPI_COMM_WORLD,&rank);
MPI_Comm_size(MPI_COMM_WORLD,&ntasks);
if (((ntasks%2) !=0) || ((LIMIT%ntasks) !=0)) {
   printf("Aquesta aplicacio requereix un nombre de tasques parell.\n");
   MPI_Finalize();
   exit(0);
}

start_time = MPI_Wtime();    /* Initialize start time */
mystart = (rank*2)+1;        /* Find my starting point - must be odd number */
stride = ntasks*2;           /* Determine stride, skipping even numbers */
pc=0;                        /* Initialize prime counter */
foundone = 0;                /* Initialize */

/********************  Tasca 0 - Inicia el procés  ********************/
if (rank == FIRST) {
   pc = 4;                  
   for (n=mystart; n<=LIMIT; n=n+stride) {
      if (isprime(n)) {
         pc++;
         foundone = n;
      }
   }
   MPI_Reduce(&pc,&pcsum,1,MPI_INT,MPI_SUM,FIRST,MPI_COMM_WORLD);
   MPI_Reduce(&foundone,&maxprime,1,MPI_INT,MPI_MAX,FIRST,MPI_COMM_WORLD);
   end_time=MPI_Wtime();
   printf("Temps total: %.2lf segonds\n",(end_time-start_time)*10);
}


/********************  Resta de tasques  ***********************/
if (rank > FIRST) {
   for (n=mystart; n<=LIMIT; n=n+stride) {
      if (isprime(n)) {
         pc++;
         foundone = n;
      }
   }
   MPI_Reduce(&pc,&pcsum,1,MPI_INT,MPI_SUM,FIRST,MPI_COMM_WORLD);
   MPI_Reduce(&foundone,&maxprime,1,MPI_INT,MPI_MAX,FIRST,MPI_COMM_WORLD);
   }

MPI_Finalize();

}
